import time
from os import makedirs
from os.path import abspath, exists, join

import matplotlib.pyplot as plt
import networkx as nx
from matplotlib.patches import Polygon

from extremitypathfinder.extremitypathfinder import PolygonEnvironment

EXPORT_RESOLUTION = 200  # dpi
EXPORT_SIZE_X = 19.0  # inch
EXPORT_SIZE_Y = 11.0  # inch

POLYGON_SETTINGS = {
    "edgecolor": "black",
    "fill": False,
    "linewidth": 1.0,
}

SHOW_PLOTS = False
PLOTTING_DIR = "all_plots"


def get_plot_name(file_name="plot"):
    return abspath(join(PLOTTING_DIR, file_name + "_" + str(time.time())[:-7] + ".png"))


def export_plot(fig, file_name):
    fig.set_size_inches(EXPORT_SIZE_X, EXPORT_SIZE_Y, forward=True)
    plt.savefig(get_plot_name(file_name), dpi=EXPORT_RESOLUTION)
    plt.close()


def mark_points(vertex_iter, **kwargs):
    try:
        coordinates = [v.tolist() for v in vertex_iter]
    except AttributeError:
        coordinates = list(vertex_iter)
    coords_zipped = list(zip(*coordinates))
    if coords_zipped:  # there might be no vertices at all
        plt.scatter(*coords_zipped, **kwargs)


def draw_edge(v1, v2, c, alpha, **kwargs):
    x1, y1 = v1
    x2, y2 = v2
    plt.plot([x1, x2], [y1, y2], color=c, alpha=alpha, **kwargs)


def draw_polygon(ax, coords, **kwargs):
    kwargs.update(POLYGON_SETTINGS)
    polygon = Polygon(coords, **kwargs)
    ax.add_patch(polygon)


def draw_boundaries(map, ax):
    # TODO outside dark grey
    # TODO fill holes light grey
    draw_polygon(ax, map.boundary_polygon)
    for h in map.holes:
        draw_polygon(ax, h, facecolor="grey", fill=True)

    mark_points(map.all_vertices, c="black", s=15)
    mark_points(map.all_extremities, c="red", s=50)


def draw_internal_graph(map: PolygonEnvironment, ax):
    graph = map.graph
    coords = map.coords
    for n in graph.nodes:
        start = coords[n]
        all_goals = [coords[i] for i in graph.neighbors(n)]
        for goal in all_goals:
            draw_edge(start, goal, c="red", alpha=0.2, linewidth=2)


def set_limits(map, ax):
    ax.set_xlim(
        (
            min(map.boundary_polygon[:, 0]) - 1,
            max(map.boundary_polygon[:, 0]) + 1,
        )
    )
    ax.set_ylim(
        (
            min(map.boundary_polygon[:, 1]) - 1,
            max(map.boundary_polygon[:, 1]) + 1,
        )
    )


def draw_path(vertex_path):
    # start, path and goal in green
    if not vertex_path:
        return
    mark_points(vertex_path, c="g", alpha=0.9, s=50)
    mark_points([vertex_path[0], vertex_path[-1]], c="g", s=100)
    v1 = vertex_path[0]
    for v2 in vertex_path[1:]:
        draw_edge(v1, v2, c="g", alpha=1.0)
        v1 = v2


def draw_loaded_map(map):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    set_limits(map, ax)
    export_plot(fig, "map_plot")
    if SHOW_PLOTS:
        plt.show()


def draw_prepared_map(map):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    draw_internal_graph(map, ax)
    set_limits(map, ax)
    export_plot(fig, "prepared_map_plot")
    if SHOW_PLOTS:
        plt.show()


def draw_with_path(map, graph: nx.DiGraph, vertex_path):
    fig, ax = plt.subplots()

    coords = map._coords_tmp
    all_nodes = graph.nodes
    draw_boundaries(map, ax)
    draw_internal_graph(map, ax)
    set_limits(map, ax)

    if len(vertex_path) > 0:
        # additionally draw:
        # new edges yellow
        start, goal = vertex_path[0], vertex_path[-1]
        goal_idx = map._idx_goal_tmp
        start_idx = map._idx_start_tmp

        if start_idx is not None:
            for n_idx in graph.neighbors(start_idx):
                n = coords[n_idx]
                draw_edge(start, n, c="y", alpha=0.7)

        if goal_idx is not None:
            # edges only run towards goal
            for n_idx in all_nodes:
                if goal_idx in graph.neighbors(n_idx):
                    n = coords[n_idx]
                    draw_edge(n, goal, c="y", alpha=0.7)

    # start, path and goal in green
    draw_path(vertex_path)

    export_plot(fig, "graph_path_plot")
    if SHOW_PLOTS:
        plt.show()


def draw_only_path(map, vertex_path, start_coordinates, goal_coordinates):
    fig, ax = plt.subplots()

    draw_boundaries(map, ax)
    set_limits(map, ax)
    draw_path(vertex_path)
    mark_points([start_coordinates, goal_coordinates], c="g", s=100)

    export_plot(fig, "path_plot")
    if SHOW_PLOTS:
        plt.show()


def draw_graph(map, graph: nx.DiGraph):
    fig, ax = plt.subplots()

    nodes = graph.nodes
    coords = map._coords_tmp
    all_nodes = [coords[i] for i in nodes]
    mark_points(all_nodes, c="black", s=30)

    for i in nodes:
        x, y = coords[i]
        neighbour_idxs = graph.neighbors(i)
        for n2_idx in neighbour_idxs:
            x2, y2 = coords[n2_idx]
            dx, dy = x2 - x, y2 - y
            plt.arrow(
                x,
                y,
                dx,
                dy,
                head_width=0.15,
                head_length=0.5,
                head_starts_at_zero=False,
                shape="full",
                length_includes_head=True,
            )

    set_limits(map, ax)

    export_plot(fig, "graph_plot")
    if SHOW_PLOTS:
        plt.show()


class PlottingEnvironment(PolygonEnvironment):
    """Extends PolygonEnvironment. In addition to the base functionality it
    plots graphs of the polygons, the visibility graph and the computed path.
    Stores all graphs in the folder defined by plotting_dir parameter."""

    def __init__(self, plotting_dir=PLOTTING_DIR):
        super().__init__()
        global PLOTTING_DIR
        PLOTTING_DIR = plotting_dir
        if not exists(plotting_dir):
            makedirs(plotting_dir)

    def store(self, *args, **kwargs):
        """In addition to storing, also plots a graph of the input polygons."""
        super().store(*args, **kwargs)
        draw_loaded_map(self)

    def prepare(self):
        """Also draws a prepared map with the computed visibility graph."""
        super().prepare()
        draw_prepared_map(self)

    def find_shortest_path(self, start_coordinates, goal_coordinates, *args, **kwargs):
        """Also draws the computed shortest path."""
        # important to not delete the temp graph! for plotting
        vertex_path, distance = super().find_shortest_path(
            start_coordinates, goal_coordinates, *args, free_space_after=False, **kwargs
        )

        draw_only_path(self, vertex_path, start_coordinates, goal_coordinates)
        if self.temp_graph:  # in some cases (e.g. direct path possible) no graph is being created!
            draw_graph(self, self.temp_graph)
            draw_with_path(self, self.temp_graph, vertex_path)
            del self.temp_graph  # free the memory

        # extract the coordinates from the path
        return vertex_path, distance



import itertools
import json
import math
import pickle
from itertools import combinations
from typing import Dict, Iterable, List, Optional, Set, Tuple

import networkx as nx
import numpy as np
import numpy.linalg

from extremitypathfinder.configs import BOUNDARY_JSON_KEY, DEFAULT_PICKLE_NAME, HOLES_JSON_KEY


def compute_repr_n_dist(np_vector: np.ndarray) -> Tuple[float, float]:
    """computing representation for the angle from the origin to a given vector

    value in [0.0 : 4.0[
    every quadrant contains angle measures from 0.0 to 1.0
    there are 4 quadrants (counter clockwise numbering)
    0 / 360 degree -> 0.0
    90 degree -> 1.0
    180 degree -> 2.0
    270 degree -> 3.0
    ...
    Useful for comparing angles without actually computing expensive trigonometrical functions
    This representation does not grow directly proportional to its represented angle,
    but it its bijective and monotonous:
    rep(p1) > rep(p2) <=> angle(p1) > angle(p2)
    rep(p1) = rep(p2) <=> angle(p1) = angle(p2)
    angle(p): counter clockwise angle between the two line segments (0,0)'--(1,0)' and (0,0)'--p
    with (0,0)' being the vector representing the origin

    :param np_vector:
    :return:
    """
    dx, dy = np_vector
    distance = math.sqrt(dx**2 + dy**2)  # l-2 norm
    if distance == 0.0:
        angle_measure = np.nan
    else:
        # 2D vector: (dx, dy) = np_vector
        dx_positive = dx >= 0
        dy_positive = dy >= 0

        if dx_positive and dy_positive:
            quadrant = 0.0
            angle_measure = dy

        elif not dx_positive and dy_positive:
            quadrant = 1.0
            angle_measure = -dx

        elif not dx_positive and not dy_positive:
            quadrant = 2.0
            angle_measure = -dy

        else:
            quadrant = 3.0
            angle_measure = dx

        # normalise angle measure to [0; 1]
        angle_measure /= distance
        angle_measure += quadrant

    return angle_measure, distance


def cmp_reps_n_distances(orig_idx: int, coords: np.ndarray) -> np.ndarray:
    coords_orig = coords[orig_idx]
    coords_translated = coords - coords_orig
    repr_n_dists = np.apply_along_axis(compute_repr_n_dist, axis=1, arr=coords_translated)
    return repr_n_dists.T


def inside_polygon(p: np.ndarray, coords: np.ndarray, border_value: bool) -> bool:
    # should return the border value for point equal to any polygon vertex
    # TODO overflow possible with large values when comparing slopes, change procedure
    # and if the point p lies on any polygon edge
    p1 = coords[-1, :]
    for p2 in coords[:]:
        if np.array_equal(p2, p):
            return border_value
        rep_p1_p, _ = compute_repr_n_dist(p1 - p)
        rep_p2_p, _ = compute_repr_n_dist(p2 - p)
        if abs(rep_p1_p - rep_p2_p) == 2.0:
            return border_value
        p1 = p2

    # regular point in polygon algorithm
    # TODO use optimised implementation
    x, y = p

    contained = False
    # the edge from the last to the first point is checked first
    i = -1
    y1 = coords[-1, 1]
    y_gt_y1 = y > y1
    for y2 in coords[:, 1]:
        y_gt_y2 = y > y2
        if y_gt_y1:
            if not y_gt_y2:
                x1 = coords[i, 0]
                x2 = coords[i + 1, 0]
                # only crossings "right" of the point should be counted
                x1GEx = x <= x1
                x2GEx = x <= x2
                # compare the slope of the line [p1-p2] and [p-p2]
                # depending on the position of p2 this determines whether the polygon edge is right or left of the point
                # to avoid expensive division the divisors (of the slope dy/dx) are brought to the other side
                # ( dy/dx > a  ==  dy > a * dx )
                if (x1GEx and x2GEx) or ((x1GEx or x2GEx) and (y2 - y) * (x2 - x1) <= (y2 - y1) * (x2 - x)):
                    contained = not contained

        else:
            if y_gt_y2:
                x1 = coords[i, 0]
                x2 = coords[i + 1, 0]
                # only crossings "right" of the point should be counted
                x1GEx = x <= x1
                x2GEx = x <= x2
                if (x1GEx and x2GEx) or ((x1GEx or x2GEx) and (y2 - y) * (x2 - x1) >= (y2 - y1) * (x2 - x)):
                    contained = not contained

        y1 = y2
        y_gt_y1 = y_gt_y2
        i += 1

    return contained


def is_within_map(p: np.ndarray, boundary: np.ndarray, holes: Iterable[np.ndarray]) -> bool:
    if not inside_polygon(p, boundary, border_value=True):
        return False
    for hole in holes:
        if inside_polygon(p, hole, border_value=False):
            return False
    return True


def no_identical_consequent_vertices(coords):
    p1 = coords[-1]
    for p2 in coords:
        # TODO adjust allowed difference: rtol, atol
        if np.array_equal(p1, p2):
            return False
        p1 = p2

    return True


def get_intersection_status(p1, p2, q1, q2):
    # return:
    #   0: no intersection
    #   1: intersection in ]p1;p2[
    # TODO support 2 remaining possibilities
    #   2: intersection directly in p1 or p2
    #   3: intersection directly in q1 or q2
    # solve the set of equations
    # (p2-p1) lambda + (p1) = (q2-q1) mu + (q1)
    #  in matrix form A x = b:
    # [(p2-p1) (q1-q2)] (lambda, mu)' = (q1-p1)
    A = np.array([p2 - p1, q1 - q2]).T
    b = np.array(q1 - p1)
    try:
        x = np.linalg.solve(A, b)
        # not crossing the line segment is considered to be ok
        # so x == 0.0 or x == 1.0 is not considered an intersection
        # assert np.allclose((p2 - p1) * x[0] + p1, (q2 - q1) * x[1] + q1)
        # assert np.allclose(np.dot(A, x), b)
        if x[0] <= 0.0 or x[1] <= 0.0 or x[0] >= 1.0 or x[1] >= 1.0:
            return 0
        # if np.all(0.0 <= x) and np.all(x <= 1.0):
        #     return 2
    except np.linalg.LinAlgError:
        # line segments are parallel (matrix is singular, set of equations is not solvable)
        return 0

    return 1


# special case of has_intersection()
def lies_behind_inner(p1: np.ndarray, p2: np.ndarray, v: np.ndarray) -> bool:
    # solve the set of equations
    # (p2-p1) lambda + (p1) = (v) mu
    #  in matrix form A x = b:
    # [(p1-p2) (v)] (lambda, mu)' = (p1)
    # because the vertex lies within the angle range between the two edge vertices
    #    (together with the other conditions on the polygons)
    #   this set of linear equations is always solvable (the matrix is regular)
    A = np.array([p1 - p2, v]).T
    b = np.array(p1)
    x = np.linalg.solve(A, b)

    # Debug:
    # try:
    #     x = np.linalg.solve(A, b)
    # except np.linalg.LinAlgError:
    #     raise ValueError
    # assert np.allclose((p2 - p1) * x[0] + p1, v * x[1])
    # assert np.allclose(np.dot(A, x), b)

    # vertices on the edge are possibly visible! ( < not <=)
    return x[1] < 1.0


def lies_behind(idx_p1: int, idx_p2: int, idx_v: int, idx_orig: int, coords: np.ndarray) -> bool:
    coords_origin = coords[idx_orig]
    coords_p1_rel = coords[idx_p1] - coords_origin
    coords_p2_rel = coords[idx_p2] - coords_origin
    coords_v_rel = coords[idx_v] - coords_origin
    return lies_behind_inner(coords_p1_rel, coords_p2_rel, coords_v_rel)


def no_self_intersection(coords):
    polygon_length = len(coords)
    # again_check = []
    for index_p1, index_q1 in combinations(range(polygon_length), 2):
        # always: index_p1 < index_q1
        if index_p1 == index_q1 - 1 or index_p1 == index_q1 + 1:
            # neighbouring edges never have an intersection
            continue
        p1, p2 = coords[index_p1], coords[(index_p1 + 1) % polygon_length]
        q1, q2 = coords[index_q1], coords[(index_q1 + 1) % polygon_length]
        intersect_status = get_intersection_status(p1, p2, q1, q2)
        if intersect_status == 1:
            return False
        # if intersect_status == 2:
        # TODO 4 different options. check which side the other edge lies on.
        # if edge changes sides this is a an intersection
        # again_check.append((p1, p2, q1, q2))
        # print(p1, p2, q1, q2)

    # TODO check for intersections across 2 edges! use computed intersection

    return True


def has_clockwise_numbering(coords: np.ndarray) -> bool:
    """tests if a polygon has clockwise vertex numbering
    approach: Sum over the edges, (x2 − x1)(y2 + y1). If the result is positive the curve is clockwise.
    from:
    https://stackoverflow.com/questions/1165647/how-to-determine-if-a-list-of-polygon-points-are-in-clockwise-order
    :param coords: the list of (x,y) coordinates representing the polygon to be tested
    :return: true if the polygon has been given in clockwise numbering
    """
    total_sum = 0.0
    p1 = coords[-1]
    for p2 in coords:
        x1, y1 = p1
        x2, y2 = p2
        total_sum += (x2 - x1) * (y2 + y1)
        p1 = p2
    return total_sum > 0


def check_polygon(polygon):
    """ensures that all the following conditions on the polygons are fulfilled:
    - must at least contain 3 vertices
    - no consequent vertices with identical coordinates in the polygons! In general might have the same coordinates
    - a polygon must not have self intersections (intersections with other polygons are allowed)
    """
    if not polygon.shape[0] >= 3:
        raise TypeError("Given polygons must at least contain 3 vertices.")
    if not polygon.shape[1] == 2:
        raise TypeError("Each point of a polygon must consist of two values (x,y).")
    if not no_identical_consequent_vertices(polygon):
        raise ValueError("Consequent vertices of a polynomial must not be identical.")
    if not no_self_intersection(polygon):
        raise ValueError("A polygon must not intersect it")


def check_data_requirements(boundary_coords: np.ndarray, list_hole_coords: List[np.ndarray]):
    """ensures that all the following conditions on the polygons are fulfilled:
        - basic polygon requirements (s. above)
        - edge numbering has to follow this convention (for easier computations):
            * outer boundary polygon: counter clockwise
            * holes: clockwise

    TODO test
    todo - polygons must not intersect each other
    TODO data rectification

    :param boundary_coords:
    :param list_hole_coords:
    :return:
    """
    check_polygon(boundary_coords)
    if has_clockwise_numbering(boundary_coords):
        raise ValueError("Vertex numbering of the boundary polygon must be counter clockwise.")
    for hole_coords in list_hole_coords:
        check_polygon(hole_coords)
        if not has_clockwise_numbering(hole_coords):
            raise ValueError("Vertex numbering of hole polygon must be clockwise.")


def find_within_range(
    vert_idx2repr: np.ndarray,
    repr1: float,
    repr2: float,
    candidate_idxs: Set[int],
    angle_range_less_180: bool,
    equal_repr_allowed: bool,
) -> Set[int]:
    """
    filters out all vertices whose representation lies within the range between
      the two given angle representations
    which range ('clockwise' or 'counter-clockwise') should be checked is determined by:
      - query angle (range) is < 180deg or not (>= 180deg)
    :param repr1:
    :param repr2:
    :param representations:
    :param angle_range_less_180: whether the angle between repr1 and repr2 is < 180 deg
    :param equal_repr_allowed: whether vertices with the same representation should also be returned
    :return:
    """
    if len(candidate_idxs) == 0:
        return set()

    repr_diff = abs(repr1 - repr2)
    if repr_diff == 0.0:
        return set()

    min_repr = min(repr1, repr2)
    max_repr = max(repr1, repr2)  # = min_angle + angle_diff

    def repr_within(r):
        # Note: vertices with the same representation will not NOT be returned!
        return min_repr < r < max_repr

    # depending on the angle the included range is clockwise or anti-clockwise
    # (from min_repr to max_val or the other way around)
    # when the range contains the 0.0 value (transition from 3.99... -> 0.0)
    # it is easier to check if a representation does NOT lie within this range
    # -> invert filter condition
    # special case: angle == 180deg
    on_line_inv = repr_diff == 2.0 and repr1 >= repr2
    # which range to filter is determined by the order of the points
    # since the polygons follow a numbering convention,
    # the 'left' side of p1-p2 always lies inside the map
    # -> filter out everything on the right side (='outside')
    inversion_condition = on_line_inv or ((repr_diff < 2.0) ^ angle_range_less_180)

    def within_filter_func(r: float) -> bool:
        repr_eq = r == min_repr or r == max_repr
        if repr_eq and equal_repr_allowed:
            return True
        if repr_eq and not equal_repr_allowed:
            return False

        res = repr_within(r)
        if inversion_condition:
            res = not res
        return res

    idxs_within = {i for i in candidate_idxs if within_filter_func(vert_idx2repr[i])}
    return idxs_within


def get_neighbour_idxs(i: int, vertex_edge_idxs: np.ndarray, edge_vertex_idxs: np.ndarray) -> Tuple[int, int]:
    edge_idx1, edge_idx2 = vertex_edge_idxs[i]
    neigh_idx1 = edge_vertex_idxs[edge_idx1, 0]
    neigh_idx2 = edge_vertex_idxs[edge_idx2, 1]
    return neigh_idx1, neigh_idx2


def skip_edge(
    node: int,
    edge2discard: int,
    candidates: Set[int],
    edge_vertex_idxs: np.ndarray,
    extremity_mask: np.ndarray,
    vertex_edge_idxs: np.ndarray,
):
    # (note: not identical, does not belong to the same polygon!)
    # mark this vertex as not visible (would otherwise add 0 distance edge in the graph)
    candidates.discard(node)
    # no points lie truly "behind" this edge as there is no "direction of sight" defined
    # <-> angle representation/range undefined for just this single edge
    # however if one considers the point neighbouring in the other direction (<-> two edges)
    # these two neighbouring edges define an invisible angle range
    # -> simply move the pointer
    v1, v2 = get_neighbour_idxs(node, vertex_edge_idxs, edge_vertex_idxs)
    range_less_180 = extremity_mask[node]
    # do not check the other neighbouring edge of vertex1 in the future (has been considered already)
    edge_idx = vertex_edge_idxs[node][edge2discard]
    return edge_idx, range_less_180, v1, v2


def find_candidates_behind(
    origin: int, v1: int, v2: int, candidates: Set[int], distances: np.ndarray, coords: np.ndarray
) -> Set[int]:
    dist_v1 = distances[v1]
    dist_v2 = distances[v2]
    max_distance = max(dist_v1, dist_v2)
    idxs_behind = set()
    # for all remaining vertices v it has to be tested if the line segment from query point (=origin) to v
    #    has an intersection with the current edge p1---p2
    for idx in candidates:
        # if a candidate is farther away from the query point than both vertices of the edge,
        #   it surely lies behind the edge
        # ATTENTION: even if a candidate is closer to the query point than both vertices of the edge,
        #   it still needs to be checked!
        dist2orig = distances[idx]
        further_away = dist2orig > max_distance
        if further_away or lies_behind(v1, v2, idx, origin, coords):
            idxs_behind.add(idx)
        # vertex lies in front of this edge
    return idxs_behind


def clean_visibles(visible_idxs: Set[int], cand_idx2repr: np.ndarray, vert_idx2dist: np.ndarray) -> Set[int]:
    # in case some vertices have the same representation, only return (link) the closest vertex
    if len(visible_idxs) <= 1:
        return visible_idxs

    cleaned = set()
    visible_idxs_sorted = sorted(visible_idxs, key=lambda i: cand_idx2repr[i])
    min_dist = np.inf
    first_idx = visible_idxs_sorted[0]
    rep_prev = cand_idx2repr[first_idx]
    selected_idx = 0
    for i in visible_idxs_sorted:
        rep = cand_idx2repr[i]
        if rep != rep_prev:
            cleaned.add(selected_idx)
            min_dist = np.inf
            rep_prev = rep

        dist = vert_idx2dist[i]
        if dist < min_dist:
            selected_idx = i
            min_dist = dist

    cleaned.add(selected_idx)
    return cleaned


def find_visible(
    origin: int,
    candidates: Set[int],
    edges_to_check: Set[int],
    coords: np.ndarray,
    representations: np.ndarray,
    distances: np.ndarray,
    extremity_mask: np.ndarray,
    edge_vertex_idxs: np.ndarray,
    vertex_edge_idxs: np.ndarray,
) -> Set[int]:
    """
    query_vertex: a vertex for which the visibility to the vertices should be checked.
        also non extremity vertices, polygon vertices and vertices with the same coordinates are allowed.
        query point also might lie directly on an edge! (angle = 180deg)
    :param candidates: the set of all vertex ids which should be checked for visibility.
        IMPORTANT: is being manipulated, so has to be a copy!
        IMPORTANT: must not contain the query vertex!
    :param edges_to_check: the set of edges which determine visibility
    :return: a set of tuples of all vertices visible from the query vertex and the corresponding distance
    """
    if len(candidates) == 0:
        return candidates

    visibles = set()
    # goal: eliminating all vertices lying 'behind' any edge
    while len(candidates) > 0:
        try:
            edge = edges_to_check.pop()
        except KeyError:
            break  # no more edges left to check

        v1, v2 = edge_vertex_idxs[edge]
        lies_on_edge = False
        range_less_180 = False

        if distances[v1] == 0.0:
            # vertex1 of the edge has the same coordinates as the query vertex
            # -> the origin lies on the edge
            lies_on_edge = True
            edge, range_less_180, v1, v2 = skip_edge(
                node=v1,
                edge2discard=0,
                candidates=candidates,
                edge_vertex_idxs=edge_vertex_idxs,
                extremity_mask=extremity_mask,
                vertex_edge_idxs=vertex_edge_idxs,
            )
            edges_to_check.discard(edge)

        elif distances[v2] == 0.0:
            # same for vertex2 of the edge
            # NOTE: it is unsupported that v1 as well as v2 have the same coordinates as the query vertex
            # (edge with length 0)
            lies_on_edge = True
            edge, range_less_180, v1, v2 = skip_edge(
                node=v2,
                edge2discard=1,
                candidates=candidates,
                edge_vertex_idxs=edge_vertex_idxs,
                extremity_mask=extremity_mask,
                vertex_edge_idxs=vertex_edge_idxs,
            )
            edges_to_check.discard(edge)

        repr1 = representations[v1]
        repr2 = representations[v2]
        repr_diff = abs(repr1 - repr2)
        if repr_diff == 2.0:
            # angle == 180deg -> on the edge
            lies_on_edge = True

        if lies_on_edge:
            # when the query vertex lies on an edge (or vertex) no behind/in front checks must be performed!
            # the neighbouring edges are visible for sure
            # attention: only add to visible set if vertex was a candidate!
            try:
                candidates.remove(v1)
                visibles.add(v1)
            except KeyError:
                pass
            try:
                candidates.remove(v2)
                visibles.add(v2)
            except KeyError:
                pass

            # all the candidates between the two vertices v1 v2 are not visible for sure
            # candidates with the same representation must not be deleted, because they can be visible!
            invisible_candidate_idxs = find_within_range(
                representations, repr1, repr2, candidates, angle_range_less_180=range_less_180, equal_repr_allowed=False
            )
            candidates.difference_update(invisible_candidate_idxs)
            continue

        # case: a 'regular' edge
        # eliminate all candidates which are blocked by the edge
        # that means inside the angle range spanned by the edge and actually behind it
        cand_idxs_tmp = candidates.copy()
        # the vertices belonging to the edge itself (its vertices) must not be checked.
        # use discard() instead of remove() to not raise an error (they might not be candidates)
        cand_idxs_tmp.discard(v1)
        cand_idxs_tmp.discard(v2)

        # for all candidate edges check if there are any candidate vertices (besides the ones belonging to the edge)
        #   within this angle range
        # the "view range" of an edge from a query point (spanned by the two vertices of the edge)
        #   is always < 180deg when the edge is not running through the query point (=180 deg)
        #  candidates with the same representation as v1 or v2 should be considered.
        #   they can be visible, but should be ruled out if they lie behind any edge!
        cand_idxs_tmp = find_within_range(
            representations, repr1, repr2, cand_idxs_tmp, angle_range_less_180=True, equal_repr_allowed=True
        )

        idxs_behind = find_candidates_behind(origin, v1, v2, cand_idxs_tmp, distances, coords)
        # vertices behind any edge are not visible
        candidates.difference_update(idxs_behind)

    # all edges have been checked
    # all remaining vertices were not concealed behind any edge and hence are visible
    visibles.update(candidates)
    return clean_visibles(visibles, representations, distances)


def find_visible_and_in_front(
    origin: int,
    nr_edges: int,
    coords: np.ndarray,
    candidates: Set[int],
    extremities: Iterable[int],
    extremity_mask: np.ndarray,
    distances: np.ndarray,
    representations: np.ndarray,
    vertex_edge_idxs: np.ndarray,
    edge_vertex_idxs: np.ndarray,
):
    # vertices all belong to a polygon
    n1, n2 = get_neighbour_idxs(origin, vertex_edge_idxs, edge_vertex_idxs)
    # ATTENTION: polygons may intersect -> neighbouring extremities must NOT be visible from each other!
    # eliminate all vertices 'behind' the query point from the candidate set
    # since the query vertex is an extremity the 'outer' angle is < 180 degree
    # then the difference between the angle representation of the two edges has to be < 2.0
    # all vertices between the angle of the two neighbouring edges ('outer side')
    #   are not visible (no candidates!)
    # ATTENTION: vertices with the same angle representation might be visible and must NOT be deleted!
    n1_repr = representations[n1]
    n2_repr = representations[n2]
    idxs_behind = find_within_range(
        representations, n1_repr, n2_repr, candidates, angle_range_less_180=True, equal_repr_allowed=False
    )
    # do not consider points found to lie behind
    candidates.difference_update(idxs_behind)
    # as shown in [1, Ch. II 4.4.2 "Property One"]: Starting from any point lying "in front of" an extremity e,
    # such that both adjacent edges are visible, one will never visit e, because everything is
    # reachable on a shorter path without e (except e itself).
    # An extremity e1 lying in the area "in front of" extremity e hence is never the next vertex
    # in a shortest path coming from e.
    # And also in reverse: when coming from e1 everything else than e itself can be reached faster
    # without visiting e.
    # -> e1 and e do not have to be connected in the graph.
    # IMPORTANT: this condition only holds for building the basic visibility graph without start and goal node!
    # When a query point (start/goal) happens to be an extremity, edges to the (visible) extremities in front
    # MUST be added to the graph!
    # Find extremities which fulfill this condition for the given query extremity
    # IMPORTANT: check all extremities here, not just current candidates
    # do not check extremities with equal coords_rel (also query extremity itself!)
    #   and with the same angle representation (those edges must not get deleted from graph!)
    idxs_in_front = find_within_range(
        representations,
        repr1=angle_rep_inverse(n1_repr),
        repr2=angle_rep_inverse(n2_repr),
        candidate_idxs=set(extremities),
        angle_range_less_180=True,
        equal_repr_allowed=False,
    )
    # do not consider points lying in front when looking for visible extremities,
    # even if they are actually be visible.
    candidates.difference_update(idxs_in_front)
    # all edges have to be checked, except the 2 neighbouring edges (handled above!)
    edge_idxs2check = set(range(nr_edges))
    edge_idxs2check.difference_update(vertex_edge_idxs[origin])
    visible_idxs = find_visible(
        origin,
        candidates,
        edge_idxs2check,
        coords,
        representations,
        distances,
        extremity_mask,
        edge_vertex_idxs,
        vertex_edge_idxs,
    )
    return idxs_in_front, visible_idxs


def compute_graph_edges(
    nr_edges: int,
    extremity_indices: Iterable[int],
    reprs_n_distances: Dict[int, np.ndarray],
    coords: np.ndarray,
    edge_vertex_idxs: np.ndarray,
    extremity_mask: np.ndarray,
    vertex_edge_idxs: np.ndarray,
) -> Dict[Tuple[int, int], float]:
    # IMPORTANT: add all extremities (even if they turn out to be dangling in the end)

    connections = {}
    for extr_ptr, origin_idx in enumerate(extremity_indices):
        vert_idx2repr, vert_idx2dist = reprs_n_distances[origin_idx]
        # optimisation: extremities are always visible to each other
        # (bi-directional relation -> undirected edges in the graph)
        #  -> do not check extremities which have been checked already
        #  (must give the same result when algorithms are correct)
        # the origin extremity itself must also not be checked when looking for visible neighbours
        candidate_idxs = set(extremity_indices[extr_ptr + 1 :])
        idxs_in_front, visible_idxs = find_visible_and_in_front(
            origin_idx,
            nr_edges,
            coords,
            candidate_idxs,
            extremity_indices,
            extremity_mask,
            vert_idx2dist,
            vert_idx2repr,
            vertex_edge_idxs,
            edge_vertex_idxs,
        )
        # "thin out" the graph:
        # remove already existing edges in the graph to the extremities in front

        for i in idxs_in_front:
            connections.pop((origin_idx, i), None)
            connections.pop((i, origin_idx), None)

        for i in visible_idxs:
            d = vert_idx2dist[i]
            connections[(origin_idx, i)] = d
            connections[(i, origin_idx)] = d

    return connections


def get_distance(n1, n2, reprs_n_distances):
    if n2 > n1:
        # Note: start and goal nodex get added last -> highest idx
        # for the lower idxs the distances to goal and start have not been computed
        # -> use the higher indices to access the distances
        tmp = n1
        n1 = n2
        n2 = tmp
    _, dists = reprs_n_distances[n1]
    distance = dists[n2]
    return distance


def find_identical(candidates: Iterable[int], reprs_n_distances: Dict[int, np.ndarray]) -> Dict[int, int]:
    # for shortest path computations all graph nodes should be unique
    # join all nodes with the same coordinates
    merging_mapping = {}
    # symmetric relation -> only consider one direction
    for n1, n2 in itertools.combinations(candidates, 2):
        dist = get_distance(n1, n2, reprs_n_distances)
        if dist == 0.0:  # same coordinates
            merging_mapping[n2] = n1

    return merging_mapping


def find_identical_single(i: int, candidates: Iterable[int], reprs_n_distances: Dict[int, np.ndarray]) -> int:
    # for shortest path computations all graph nodes should be unique
    # join all nodes with the same coordinates
    # symmetric relation -> only consider one direction
    for n in candidates:
        if i == n:
            continue
        dist = get_distance(i, n, reprs_n_distances)
        if dist == 0.0:  # same coordinates
            return n
    return i


def compute_graph(
    nr_edges: int,
    extremity_indices: Iterable[int],
    reprs_n_distances: Dict[int, np.ndarray],
    coords: np.ndarray,
    edge_vertex_idxs: np.ndarray,
    extremity_mask: np.ndarray,
    vertex_edge_idxs: np.ndarray,
) -> nx.DiGraph:
    edges = compute_graph_edges(
        nr_edges,
        extremity_indices,
        reprs_n_distances,
        coords,
        edge_vertex_idxs,
        extremity_mask,
        vertex_edge_idxs,
    )

    graph = nx.DiGraph()
    # IMPORTANT: add all extremities (even if they turn out to be dangling in the end),
    # adding start and goal nodes at query time might connect them!
    graph.add_nodes_from(extremity_indices)
    for (start, goal), dist in edges.items():
        graph.add_edge(start, goal, weight=dist)

    merge_mapping = find_identical(graph.nodes, reprs_n_distances)
    if len(merge_mapping) > 0:
        nx.relabel_nodes(graph, merge_mapping, copy=False)

    return graph


def try_extraction(json_data, key):
    try:
        extracted_data = json_data[key]
    except KeyError as e:
        raise ValueError(f"The expected key {key} was not found in the JSON file:\n{e}")
    return extracted_data


def convert2polygon(json_list):
    return [tuple(coord_pair_list) for coord_pair_list in json_list]


def read_json(path2json_file):
    """
    Parse data from a JSON file and save as lists of tuples for both boundary and holes.
    NOTE: The format of the JSON file is explained in the command line script (argparse definition)

    :param path2json_file: The path to the input json file
    :return: The parsed lists of boundaries and holes
    """
    # parse data from the input file
    with open(path2json_file, "r") as json_file:
        json_data = json_file.read()
    json_loaded = json.loads(json_data)
    boundary_data = try_extraction(json_loaded, BOUNDARY_JSON_KEY)
    holes_data = try_extraction(json_loaded, HOLES_JSON_KEY)
    boundary_coordinates = convert2polygon(boundary_data)
    list_of_holes = [convert2polygon(hole_data) for hole_data in holes_data]
    return boundary_coordinates, list_of_holes


def convert_gridworld(size_x: int, size_y: int, obstacle_iter: iter, simplify: bool = True) -> (list, list):
    """
    prerequisites: grid world must not have non-obstacle cells which are surrounded by obstacles
    ("single white cell in black surrounding" = useless for path planning)
    :param size_x: the horizontal grid world size
    :param size_y: the vertical grid world size
    :param obstacle_iter: an iterable of coordinate pairs (x,y) representing blocked grid cells (obstacles)
    :param simplify: whether the polygons should be simplified or not. reduces edge amount, allow diagonal edges
    :return: an boundary polygon (counter clockwise numbering) and a list of hole polygons (clockwise numbering)
    NOTE: convert grid world into polygons in a way that coordinates coincide with grid!
        -> no conversion of obtained graphs needed!
        the origin of the polygon coordinate system is (-0.5,-0.5) in the grid cell system (= corners of the grid world)
    """

    assert size_x > 0 and size_y > 0

    if len(obstacle_iter) == 0:
        # there are no obstacles. return just the simple boundary rectangle
        return [np.array(x, y) for x, y in [(0, 0), (size_x, 0), (size_x, size_y), (0, size_y)]], []

    # convert (x,y) into np.arrays
    # obstacle_iter = [np.array(o) for o in obstacle_iter]
    obstacle_iter = np.array(obstacle_iter)

    def within_grid(pos):
        return 0 <= pos[0] < size_x and 0 <= pos[1] < size_y

    def is_equal(pos1, pos2):
        return np.all(pos1 == pos2)

    def pos_in_iter(pos, iter):
        for i in iter:
            if is_equal(pos, i):
                return True
        return False

    def is_obstacle(pos):
        return pos_in_iter(pos, obstacle_iter)

    def is_blocked(pos):
        return not within_grid(pos) or is_obstacle(pos)

    def is_unblocked(pos):
        return within_grid(pos) and not is_obstacle(pos)

    def find_start(start_pos, boundary_detect_fct, **kwargs):
        # returns the lowest and leftmost unblocked grid cell from the start position
        # for which the detection function evaluates to True
        start_x, start_y = start_pos
        for y in range(start_y, size_y):
            for x in range(start_x, size_x):
                pos = np.array([x, y])
                if boundary_detect_fct(pos, **kwargs):
                    return pos

    # north, east, south, west
    directions = np.array([[0, 1], [1, 0], [0, -1], [-1, 0]], dtype=int)
    # the correct offset to determine where nodes should be added.
    offsets = np.array([[0, 1], [1, 1], [1, 0], [0, 0]], dtype=int)

    def construct_polygon(start_pos, boundary_detect_fct, cntr_clockwise_wanted: bool):
        current_pos = start_pos.copy()
        # (at least) the west and south are blocked
        #   -> there has to be a polygon node at the current position (bottom left corner of the cell)
        edge_list = [start_pos]
        forward_index = 0  # start with moving north
        forward_vect = directions[forward_index]
        left_index = (forward_index - 1) % 4
        # left_vect = directions[(forward_index - 1) % 4]
        just_turned = True

        # follow the border between obstacles and free cells ("wall") until one
        # reaches the start position again
        while True:
            # left has to be checked first
            # do not check if just turned left or right (-> the left is blocked for sure)
            # left_pos = current_pos + left_vect
            if not (just_turned or boundary_detect_fct(current_pos + directions[left_index])):
                # print('< turn left')
                forward_index = left_index
                left_index = (forward_index - 1) % 4
                forward_vect = directions[forward_index]
                just_turned = True

                # add a new node at the correct position
                # decrease the index first!
                edge_list.append(current_pos + offsets[forward_index])

                # move forward (previously left, there is no obstacle)
                current_pos += forward_vect
            else:
                forward_pos = current_pos + forward_vect
                if boundary_detect_fct(forward_pos):
                    node_pos = current_pos + offsets[forward_index]
                    # there is a node at the bottom left corner of the start position (offset= (0,0) )
                    if is_equal(node_pos, start_pos):
                        # check and terminate if this node does already exist
                        break

                    # add a new node at the correct position
                    edge_list.append(node_pos)
                    # print('> turn right')
                    left_index = forward_index
                    forward_index = (forward_index + 1) % 4
                    forward_vect = directions[forward_index]
                    just_turned = True
                    # print(direction_index,forward_vect,just_turned,edge_list,)
                else:
                    # print('^ move forward')
                    current_pos += forward_vect
                    just_turned = False

        if cntr_clockwise_wanted:
            # make edge numbering counter clockwise!
            edge_list.reverse()
        return np.array(edge_list, dtype=float)

    # build the boundary polygon
    # start at the lowest and leftmost unblocked grid cell
    start_pos = find_start(start_pos=(0, 0), boundary_detect_fct=is_unblocked)
    # print(start_pos+directions[3])
    # raise ValueError
    boundary_edges = construct_polygon(start_pos, boundary_detect_fct=is_blocked, cntr_clockwise_wanted=True)

    if simplify:
        # TODO
        raise NotImplementedError()

    # detect which of the obstacles have to be converted into holes
    # just the obstacles inside the boundary polygon are part of holes
    # shift coordinates by +(0.5,0.5) for correct detection
    # the border value does not matter here

    def get_unchecked_obstacles(obstacles: Iterable, poly: np.ndarray, required_val: bool = True) -> List:
        unchecked_obstacles = []
        for o in obstacles:
            p = o + 0.5
            if inside_polygon(p, poly, border_value=True) == required_val:
                unchecked_obstacles.append(o)

        return unchecked_obstacles

    unchecked_obstacles = get_unchecked_obstacles(obstacle_iter, boundary_edges)

    hole_list = []
    while len(unchecked_obstacles) > 0:
        start_pos = find_start(start_pos=(0, 0), boundary_detect_fct=pos_in_iter, iter=unchecked_obstacles)
        hole = construct_polygon(start_pos, boundary_detect_fct=is_unblocked, cntr_clockwise_wanted=False)

        # detect which of the obstacles still do not belong to any hole:
        # delete the obstacles which are included in the just constructed hole
        unchecked_obstacles = get_unchecked_obstacles(unchecked_obstacles, hole, required_val=False)

        if simplify:
            # TODO
            pass

        hole_list.append(hole)

    return boundary_edges, hole_list


def angle_rep_inverse(repr: Optional[float]) -> Optional[float]:
    if repr is None:
        repr_inv = None
    else:
        repr_inv = (repr + 2.0) % 4.0
    return repr_inv


def compute_extremity_idxs(coordinates: np.ndarray) -> List[int]:
    """identify all protruding points = vertices with an inside angle of > 180 degree ('extremities')
    expected edge numbering:
        outer boundary polygon: counter clockwise
        holes: clockwise

    basic idea:
      - translate the coordinate system to have p2 as origin
      - compute the angle representations of both vectors representing the edges
      - "rotate" the coordinate system (equal to deducting) so that the p1p2 representation is 0
      - check in which quadrant the p2p3 representation lies
    %4 because the quadrant has to be in [0,1,2,3] (representation in [0:4[)
    if the representation lies within quadrant 0 or 1 (<2.0), the inside angle
      (for boundary polygon inside, for holes outside) between p1p2p3 is > 180 degree
    then p2 = extremity
    :param coordinates:
    :return:
    """
    nr_coordinates = len(coordinates)
    extr_idxs = []
    p1 = coordinates[-2]
    p2 = coordinates[-1]
    for i, p3 in enumerate(coordinates):
        # since consequent vertices are not permitted to be equal,
        #   the angle representation of the difference is well defined
        diff_p3_p2 = p3 - p2
        diff_p1_p2 = p1 - p2
        repr_p3_p2, _ = compute_repr_n_dist(diff_p3_p2)
        repr_p1_p2, _ = compute_repr_n_dist(diff_p1_p2)
        rep_diff = repr_p3_p2 - repr_p1_p2
        if rep_diff % 4.0 < 2.0:  # inside angle > 180 degree
            # p2 is an extremity
            idx_p2 = (i - 1) % nr_coordinates
            extr_idxs.append(idx_p2)

        # move to the next point
        p1 = p2
        p2 = p3
    return extr_idxs


def load_pickle(path=DEFAULT_PICKLE_NAME):
    print("loading map from:", path)
    with open(path, "rb") as f:
        return pickle.load(f)



"""
TODO test find_visible(), ...
TODO test if relation is really bidirectional (y in find_visible(x,y) <=> x in find_visible(y,x))
TODO test input data validation
"""
from os.path import abspath, join, pardir
from typing import Dict, Set

import numpy as np
import pytest

from extremitypathfinder import PolygonEnvironment
from extremitypathfinder.utils import (
    clean_visibles,
    compute_extremity_idxs,
    compute_repr_n_dist,
    has_clockwise_numbering,
    inside_polygon,
    read_json,
)
from tests.helpers import proto_test_case


def test_inside_polygon():
    for border_value in [True, False]:

        def test_fct(input):
            polygon_test_case = np.array([(-1.0, -1.0), (1.0, -1.0), (1.0, 1.0), (-1.0, 1.0)])
            p = np.array(input, dtype=float)
            return inside_polygon(p, polygon_test_case, border_value)

        p_test_cases = [
            # (x,y),
            # inside
            (0.0, 0.0),
            # # outside
            (-2.0, 2.0),
            (0, 2.0),
            (2.0, 2.0),
            (-2.0, 0),
            (2.0, 0),
            (-2.0, -2.0),
            (0, -2.0),
            (2.0, -2.0),
            # on the line test cases
            (-1.0, -1.0),
            (1.0, -1.0),
            (1.0, 1.0),
            (-1.0, 1.0),
            (0.0, 1),
            (0, -1),
            (1, 0),
            (-1, 0),
        ]
        expected_results = [
            True,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            False,
            # on the line test cases
            border_value,
            border_value,
            border_value,
            border_value,
            border_value,
            border_value,
            border_value,
            border_value,
        ]

        proto_test_case(list(zip(p_test_cases, expected_results)), test_fct)


def test_read_json():
    path2json_file = abspath(join(__file__, pardir, pardir, "example.json"))
    boundary_coordinates, list_of_holes = read_json(path2json_file)
    assert len(boundary_coordinates) == 5
    assert len(boundary_coordinates[0]) == 2
    assert len(list_of_holes) == 2
    first_hole = list_of_holes[0]
    assert len(first_hole) == 4
    assert len(first_hole[0]) == 2
    environment = PolygonEnvironment()
    environment.store(boundary_coordinates, list_of_holes, validate=True)


@pytest.mark.parametrize(
    "visible_idxs, cand_idx2repr, vert_idx2dist, expected",
    [
        (set(), {}, {}, set()),
        ({0}, {0: 0.0}, {0: 0.0}, {0}),
        # different repr -> keep both
        ({0, 1}, {0: 0.0, 1: 1.0}, {0: 0.0, 1: 0.0}, {0, 1}),
        ({0, 1}, {0: 0.5, 1: 1.0}, {0: 0.0, 1: 1.0}, {0, 1}),
        ({0, 1}, {0: 0.5, 1: 1.0}, {0: 1.0, 1: 1.0}, {0, 1}),
        # same repr -> keep one the one with the lower dist
        ({0, 1}, {0: 0.0, 1: 0.0}, {0: 0.0, 1: 1.0}, {0}),
        ({0, 1}, {0: 0.0, 1: 0.0}, {0: 0.0, 1: 1.1}, {0}),
        ({0, 1}, {0: 0.0, 1: 0.0}, {0: 1.0, 1: 0.0}, {1}),
        ({0, 1}, {0: 0.0, 1: 0.0}, {0: 1.1, 1: 0.0}, {1}),
    ],
)
def test_clean_visible_idxs(
    visible_idxs: Set[int], cand_idx2repr: Dict[int, float], vert_idx2dist: Dict[int, float], expected: Set[int]
):
    res = clean_visibles(visible_idxs, cand_idx2repr, vert_idx2dist)
    assert res == expected


@pytest.mark.parametrize(
    "coords, expected",
    [
        (
            [
                (0, 0),
                (10, 0),
                (9, 5),
                (10, 10),
                (0, 10),
            ],
            {2},
        ),
        (
            [
                (0, 0),
                (10, 0),
                (10, 10),
                (0, 10),
            ],
            set(),
        ),
        (
            [
                (0, 0),
                (-2, -2),
                (-3, -2.5),
                (-3, -4),
                (2, -3),
                (1, 2.5),
                (0, -1),
            ],
            {6, 1},
        ),
    ],
)
def test_compute_extremity_idxs(coords, expected):
    coords = np.array(coords)
    res = compute_extremity_idxs(coords)
    assert set(res) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        # clockwise numbering!
        ([(3.0, 7.0), (5.0, 9.0), (5.0, 7.0)], True),
        ([(3.0, 7.0), (5.0, 9.0), (4.5, 7.0), (5.0, 4.0)], True),
        ([(0.0, 0.0), (0.0, 1.0), (1.0, 0.0)], True),
        # # counter clockwise edge numbering!
        ([(0.0, 0.0), (1.0, 0.0), (0.0, 1.0)], False),
        ([(0.0, 0.0), (10.0, 0.0), (10.0, 10.0), (0.0, 10.0)], False),
        ([(0.0, 0.0), (10.0, 0.0), (10.0, 5.0), (10.0, 10.0), (0.0, 10.0)], False),
        ([(0.0, 0.0), (10.0, 0.0), (9.0, 5.0), (10.0, 10.0), (0.0, 10.0)], False),
    ],
)
def test_clockwise_numering(input, expected):
    def clockwise_test_fct(input):
        return has_clockwise_numbering(np.array(input))

    assert clockwise_test_fct(input) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ([0.0, -5.0], (3.0, 5.0)),
        ([-5.0, 0.0], (2.0, 5.0)),
        ([-1.0, 0.0], (2.0, 1.0)),
        ([1.0, 0.0], (0.0, 1.0)),
        ([0.0, 1.0], (1.0, 1.0)),
        ([-6.0, -5.0], (2.64018439966448, 7.810249675906654)),
        ([-5.0, -6.0], (2.768221279597376, 7.810249675906654)),
    ],
)
def test_compute_repr_n_dist(input, expected):
    def test_fct(input):
        return compute_repr_n_dist(np.array(input))

    assert test_fct(input) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ([1.0, 0.0], 0.0),
        ([0.0, 1.0], 1.0),
        ([-1.0, 0.0], 2.0),
        ([0.0, -1.0], 3.0),
        ([2.0, 0.0], 0.0),
        ([0.0, 2.0], 1.0),
        ([-2.0, 0.0], 2.0),
        ([0.0, -2.0], 3.0),
    ],
)
def test_angle_representation(input, expected):
    def func(input):
        repr, dist = compute_repr_n_dist(np.array(input))
        return repr

    assert func(input) == expected


@pytest.mark.parametrize(
    "input, expected",
    [
        ([1.0, 0.0], 0.0),
        ([0.0, 1.0], 1.0),
        ([-1.0, 0.0], 2.0),
        ([0.0, -1.0], 3.0),
        ([2.0, 0.0], 0.0),
        ([0.0, 2.0], 1.0),
        ([-2.0, 0.0], 2.0),
        ([0.0, -2.0], 3.0),
        ([1.0, 1.0], 0.0),
        ([-1.0, 1.0], 1.0),
        ([-1.0, -1.0], 2.0),
        ([1.0, -1.0], 3.0),
        ([1.0, 0.00001], 0.0),
        ([0.00001, 1.0], 0.0),
        ([-1.0, 0.00001], 1.0),
        ([0.00001, -1.0], 3.0),
        ([1.0, -0.00001], 3.0),
        ([-0.00001, 1.0], 1.0),
        ([-1.0, -0.00001], 2.0),
        ([-0.00001, -1.0], 2.0),
    ],
)
def test_angle_repr_quadrant(input, expected):
    def func(input):
        repr, dist = compute_repr_n_dist(np.array(input))
        return repr

    res = func(input)
    assert res >= expected
    assert res < expected + 1



import pytest

from extremitypathfinder.extremitypathfinder import PolygonEnvironment
from extremitypathfinder.plotting import PlottingEnvironment
from tests.test_cases import (
    GRID_ENV_PARAMS,
    INVALID_DESTINATION_DATA,
    OVERLAP_POLY_ENV_PARAMS,
    POLY_ENV_PARAMS,
    SEPARATED_ENV,
    TEST_DATA_GRID_ENV,
    TEST_DATA_OVERLAP_POLY_ENV,
    TEST_DATA_POLY_ENV,
    TEST_DATA_SEPARATE_ENV,
)

# PLOT_TEST_RESULTS = True
PLOT_TEST_RESULTS = False
TEST_PLOT_OUTPUT_FOLDER = "plots"

if PLOT_TEST_RESULTS:
    print("plotting test environment enabled.")
    ENVIRONMENT_CLASS = PlottingEnvironment
    CONSTRUCTION_KWARGS = {"plotting_dir": TEST_PLOT_OUTPUT_FOLDER}
else:
    ENVIRONMENT_CLASS = PolygonEnvironment
    CONSTRUCTION_KWARGS = {}


# TODO pytest parameterize


def try_test_cases(environment, test_cases):
    def validate(start_coordinates, goal_coordinates, expected_output):
        output = environment.find_shortest_path(start_coordinates, goal_coordinates)
        path, length = output
        assert type(path) is list
        expected_path, expected_length = expected_output
        if expected_length is None:
            correct_result = length is None and path == expected_path
        else:
            correct_result = path == expected_path and length == pytest.approx(expected_length)
        if correct_result:
            status_str = "OK"
        else:
            status_str = "XX"
        print(f"{status_str} input: {(start_coordinates, goal_coordinates)} ")
        assert correct_result, f"unexpected result (path, length): got {output} instead of {expected_output} "

    print("testing if path and distance are correct:")
    for ((start_coordinates, goal_coordinates), expected_output) in test_cases:
        validate(start_coordinates, goal_coordinates, expected_output)
        # automatically test reversed!
        path, length = expected_output
        expected_output_reversed = list(reversed(path)), length
        validate(goal_coordinates, start_coordinates, expected_output_reversed)


def test_grid_env():
    grid_env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)

    grid_env.store_grid_world(*GRID_ENV_PARAMS, simplify=False, validate=False)
    nr_extremities = len(grid_env.all_extremities)
    assert nr_extremities == 17, "extremities do not get detected correctly!"
    grid_env.prepare()
    nr_graph_nodes = len(grid_env.graph.nodes)
    assert nr_graph_nodes == 16, "identical nodes should get joined in the graph!"

    # test if points outside the map are being rejected
    for start_coordinates, goal_coordinates in INVALID_DESTINATION_DATA:
        with pytest.raises(ValueError):
            grid_env.find_shortest_path(start_coordinates, goal_coordinates)

    print("testing grid environment")
    try_test_cases(grid_env, TEST_DATA_GRID_ENV)

    # when the deep copy mechanism works correctly
    # even after many queries the internal graph should have the same structure as before
    # otherwise the temporarily added vertices during a query stay stored
    nr_graph_nodes = len(grid_env.graph.nodes)
    # TODO
    # assert nr_graph_nodes == 16, "the graph should stay unchanged by shortest path queries!"

    nr_nodes_env1_old = len(grid_env.graph.nodes)


def test_poly_env():
    poly_env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)
    poly_env.store(*POLY_ENV_PARAMS, validate=True)
    nr_exp_extremities = 4
    assert (
        len(list(poly_env.all_extremities)) == nr_exp_extremities
    ), f"the environment should detect all {nr_exp_extremities} extremities!"
    poly_env.prepare()
    nr_nodes_env2 = len(poly_env.graph.nodes)
    # TODO
    # assert nr_nodes_env2 == nr_exp_extremities, (
    #     f"the visibility graph should store all {nr_exp_extremities} extremities {list(poly_env.all_extremities)}!"
    #     f"\n found: {poly_env.graph.nodes}"
    # )

    # nr_nodes_env1_new = len(grid_env.graph.nodes)
    # assert (
    #     nr_nodes_env1_new == nr_nodes_env1_old
    # ), "node amount of an grid_env should not change by creating another grid_env!"
    # assert grid_env.graph is not poly_env.graph, "different environments share the same graph object"
    # assert (
    #     grid_env.graph.nodes is not poly_env.graph.nodes
    # ), "different environments share the same set of nodes"

    print("\ntesting polygon environment")
    try_test_cases(poly_env, TEST_DATA_POLY_ENV)

    # TODO test: When the paths should be blocked, use a single polygon with multiple identical
    #   vertices instead (also allowed?! change data requirements in doc!).

    # TODO test graph construction
    # when two nodes have the same angle representation there should only be an edge to the closer node!
    # test if property 1 is being properly exploited
    # (extremities lying in front of each other need not be connected)


def test_overlapping_polygon():
    overlap_poly_env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)
    overlap_poly_env.store(*OVERLAP_POLY_ENV_PARAMS)
    overlap_poly_env.prepare()
    print("\ntesting polygon environment with overlapping polygons")
    try_test_cases(overlap_poly_env, TEST_DATA_OVERLAP_POLY_ENV)


def test_separated_environment():
    env = ENVIRONMENT_CLASS(**CONSTRUCTION_KWARGS)
    env.store(*SEPARATED_ENV)
    env.prepare()
    print("\ntesting polygon environment with two separated areas")
    try_test_cases(env, TEST_DATA_SEPARATE_ENV)



from typing import Iterable, List, Optional, Tuple, Union

import numpy as np

COORDINATE_TYPE = Tuple[float, float]
PATH_TYPE = List[COORDINATE_TYPE]
LENGTH_TYPE = Optional[float]
INPUT_NUMERICAL_TYPE = Union[float, int]
InputCoords = Tuple[INPUT_NUMERICAL_TYPE, INPUT_NUMERICAL_TYPE]
OBSTACLE_ITER_TYPE = Iterable[InputCoords]
INPUT_COORD_LIST_TYPE = Union[np.ndarray, List]
DEFAULT_PICKLE_NAME = "environment.pickle"

# command line interface
# json data input format:
BOUNDARY_JSON_KEY = "boundary"
HOLES_JSON_KEY = "holes"



import subprocess
from pathlib import Path
from typing import List

import pytest

PROJECT_DIR = Path(__file__).parent.parent
EXAMPLE_FILE = PROJECT_DIR / "example.json"


@pytest.mark.parametrize(
    "cmd",
    [
        f"extremitypathfinder {EXAMPLE_FILE} -s 2.5 3.2 -g 7.9 6.8",
    ],
)
def test_main(cmd: List[str]):
    res = subprocess.getoutput(cmd)
    assert not res.endswith("command not found"), "package not installed"
    splits = res.split(" ")
    length = float(splits[-1])
    print("length:", length)
    path = " ".join(splits[:-1])
    print("path:", path)



import argparse

from extremitypathfinder import PolygonEnvironment
from extremitypathfinder.configs import BOUNDARY_JSON_KEY, HOLES_JSON_KEY
from extremitypathfinder.utils import read_json

JSON_HELP_MSG = (
    "path to the JSON file to be read. "
    f'The JSON file must have 2 keys: "{BOUNDARY_JSON_KEY}" and "{HOLES_JSON_KEY}".'
    'Example: (also see "example.json")'
    '{ "boundary":[[0.0, 0.0], [10.0, 0.0],[9.0, 5.0], [10.0, 10.0], [0.0, 10.0]],'
    '"holes": [[[3.0, 7.0], [5.0, 9.0], [4.5, 7.0], [5.0, 4.0]],'
    "[[1.0, 2.0], [2.0, 2.0], [2.0, 1.0], [1.0, 1.0]]] }"
    f' The value assigned to the "{BOUNDARY_JSON_KEY}" key must be list containing '
    "a list of length 2 for each coordinate pair."
    f' The value assigned to the "{HOLES_JSON_KEY}" key must be '
    "a list of possibly multiple hole polygon lists."
)

# environment = PlottingEnvironment()
environment = PolygonEnvironment()


def main():
    parser = argparse.ArgumentParser(description="parse extremitypathfinder parameters")
    parser.add_argument("path2json_file", type=str, help=JSON_HELP_MSG)
    parser.add_argument(
        "-s",
        "--start",
        nargs=2,
        type=float,
        required=True,
        help="the start coordinates given as two float values, e.g. <2.5 3.2>",
    )
    parser.add_argument(
        "-g",
        "--goal",
        nargs=2,
        type=float,
        required=True,
        help="the goal coordinates given as two float values, e.g. <7.9 6.8>",
    )
    # TODO grid input requires different json data format
    # parser.add_argument('-g', '--grid', type=bool, help='weather the input data is a specifying a grid')
    parsed_args = parser.parse_args()  # Takes input from sys.argv

    # Parse JSON file
    boundary_coordinates, list_of_holes = read_json(parsed_args.path2json_file)

    # Parse tuples from the input arguments (interpreted as lists):
    start_coordinates = tuple(parsed_args.start)
    goal_coordinates = tuple(parsed_args.goal)

    # Execute search for the given parameters
    environment.store(boundary_coordinates, list_of_holes, validate=False)
    environment.prepare()
    path, distance = environment.find_shortest_path(start_coordinates, goal_coordinates)
    print(path, distance)



import pickle
from typing import Dict, Iterable, List, Optional, Set, Tuple

import networkx as nx
import numpy as np

from extremitypathfinder.configs import (
    DEFAULT_PICKLE_NAME,
    INPUT_COORD_LIST_TYPE,
    LENGTH_TYPE,
    OBSTACLE_ITER_TYPE,
    PATH_TYPE,
    InputCoords,
)
from extremitypathfinder.utils import (
    check_data_requirements,
    cmp_reps_n_distances,
    compute_extremity_idxs,
    compute_graph,
    convert_gridworld,
    find_identical_single,
    find_visible,
    get_distance,
    is_within_map,
)


class PolygonEnvironment:
    """Class allowing to use polygons to represent "2D environments" and use them for path finding.

    Keeps a "loaded" and prepared environment for consecutive path queries.
    Internally uses a visibility graph optimised for shortest path finding.
    General approach and some optimisations theoretically described in:
    [1] Vinther, Anders Strand-Holm, Magnus Strand-Holm Vinther, and Peyman Afshani.
    "`Pathfinding in Two-dimensional Worlds
    <https://www.cs.au.dk/~gerth/advising/thesis/anders-strand-holm-vinther_magnus-strand-holm-vinther.pdf>`__"

    TODO document parameters
    """

    nr_edges: int
    prepared: bool = False
    holes: List[np.ndarray]
    extremity_indices: List[int]
    reprs_n_distances: Dict[int, np.ndarray]
    graph: nx.DiGraph
    # TODO
    temp_graph: Optional[nx.DiGraph] = None  # for storing and plotting the graph during a query
    boundary_polygon: np.ndarray
    coords: np.ndarray
    edge_vertex_idxs: np.ndarray
    extremity_mask: np.ndarray
    vertex_edge_idxs: np.ndarray

    @property
    def nr_edges(self) -> int:
        return self.nr_vertices

    @property
    def all_extremities(self) -> List[Tuple]:
        coords = self.coords
        return [tuple(coords[i]) for i in self.extremity_indices]

    @property
    def all_vertices(self) -> List[Tuple]:
        coords = self.coords
        return [tuple(coords[i]) for i in range(self.nr_vertices)]

    def store(
        self,
        boundary_coordinates: INPUT_COORD_LIST_TYPE,
        list_of_hole_coordinates: INPUT_COORD_LIST_TYPE,
        validate: bool = False,
    ):
        """saves the passed input polygons in the environment

        .. note:: the passed polygons must meet these requirements:

            * given as numpy or python array of coordinate tuples: ``[(x1,y1), (x2,y2,)...]``
            * no repetition of the first point at the end
            * at least 3 vertices (no single points or lines allowed)
            * no consequent vertices with identical coordinates in the polygons (same coordinates allowed)
            * no self intersections
            * edge numbering has to follow these conventions: boundary polygon counter clockwise, holes clockwise

        :param boundary_coordinates: array of coordinates with counter clockwise edge numbering
        :param list_of_hole_coordinates: array of coordinates with clockwise edge numbering
        :param validate: whether the requirements of the data should be tested

        :raises AssertionError: when validate=True and the input is invalid.
        """
        self.prepared = False
        # loading the map
        boundary_coordinates = np.array(boundary_coordinates, dtype=float)
        list_of_hole_coordinates = [np.array(hole_coords, dtype=float) for hole_coords in list_of_hole_coordinates]
        if validate:
            check_data_requirements(boundary_coordinates, list_of_hole_coordinates)

        self.boundary_polygon = boundary_coordinates

        # IMPORTANT: make a copy of the list instead of linking to the same list (python!)
        self.holes = list_of_hole_coordinates

        list_of_polygons = [boundary_coordinates] + list_of_hole_coordinates
        nr_total_pts = sum(map(len, list_of_polygons))
        vertex_edge_idxs = np.empty((nr_total_pts, 2), dtype=int)
        # TODO required? inverse of the other. get_neighbours function
        edge_vertex_idxs = np.empty((nr_total_pts, 2), dtype=int)
        edge_idx = 0
        offset = 0
        extremity_idxs = set()
        for poly in list_of_polygons:
            poly_extr_idxs = compute_extremity_idxs(poly)
            poly_extr_idxs = {i + offset for i in poly_extr_idxs}
            extremity_idxs |= poly_extr_idxs

            nr_coords = len(poly)
            v1 = -1 % nr_coords
            # TODO col 1 is just np.arange?!
            for v2 in range(nr_coords):
                v1_idx = v1 + offset
                v2_idx = v2 + offset
                edge_vertex_idxs[edge_idx, 0] = v1_idx
                edge_vertex_idxs[edge_idx, 1] = v2_idx
                vertex_edge_idxs[v1_idx, 1] = edge_idx
                vertex_edge_idxs[v2_idx, 0] = edge_idx
                # move to the next vertex/edge
                v1 = v2
                edge_idx += 1

            offset = edge_idx

        # assert edge_idx == nr_total_pts

        coords = np.concatenate(list_of_polygons, axis=0)
        # Attention: only consider extremities that are actually within the map
        extremity_idxs = [i for i in extremity_idxs if self.within_map(coords[i])]

        mask = np.full(nr_total_pts, False, dtype=bool)
        for i in extremity_idxs:
            mask[i] = True

        self.nr_vertices = nr_total_pts
        # start and goal points will be stored after all polygon coordinates
        self.idx_start = nr_total_pts
        self.idx_goal = nr_total_pts + 1
        self.edge_vertex_idxs = edge_vertex_idxs
        self.vertex_edge_idxs = vertex_edge_idxs
        self.coords = coords
        self.extremity_indices = extremity_idxs
        self.extremity_mask = mask
        self.reprs_n_distances = {i: cmp_reps_n_distances(i, coords) for i in extremity_idxs}

    def store_grid_world(
        self,
        size_x: int,
        size_y: int,
        obstacle_iter: OBSTACLE_ITER_TYPE,
        simplify: bool = True,
        validate: bool = False,
    ):
        """Convert a grid-like into a polygon environment and save it

        Prerequisites: grid world must not have single non-obstacle cells which are surrounded by obstacles
        ("white cells in black surrounding" = useless for path planning)

        :param size_x: the horizontal grid world size
        :param size_y: the vertical grid world size
        :param obstacle_iter: an iterable of coordinate pairs (x,y) representing blocked grid cells (obstacles)
        :param validate: whether the input should be validated
        :param simplify: whether the polygons should be simplified or not. reduces edge amount, allow diagonal edges
        """
        boundary_coordinates, list_of_hole_coordinates = convert_gridworld(size_x, size_y, obstacle_iter, simplify)
        self.store(boundary_coordinates, list_of_hole_coordinates, validate)

    def export_pickle(self, path: str = DEFAULT_PICKLE_NAME):
        print("storing map class in:", path)
        with open(path, "wb") as f:
            pickle.dump(self, f)
        print("done.\n")

    def prepare(self):
        """Computes a visibility graph optimized (=reduced) for path planning and stores it

        Computes all directly reachable extremities based on visibility and their distance to each other
        pre-procesing of the map. pre-computation for faster shortest path queries
        optimizes graph further at construction time

        NOTE: initialise the graph with all extremities.
        even if a node has no edges (visibility to other extremities, dangling node),
        it must still be included!

        .. note::
            Multiple polygon vertices might have identical coords_rel.
            They must be treated as distinct vertices here, since their attached edges determine visibility.
            In the created graph however, these nodes must be merged at the end to avoid ambiguities!

        .. note::
            Pre computing the shortest paths between all directly reachable extremities
            and storing them in the graph would not be an advantage, because then the graph is fully connected.
            A* would visit every node in the graph at least once (-> disadvantage!).
        """
        if self.prepared:
            raise ValueError("this environment is already prepared. load new polygons first.")

        self.graph = compute_graph(
            self.nr_edges,
            self.extremity_indices,
            self.reprs_n_distances,
            self.coords,
            self.edge_vertex_idxs,
            self.extremity_mask,
            self.vertex_edge_idxs,
        )
        self.prepared = True

    def within_map(self, coords: np.ndarray) -> bool:
        """checks if the given coordinates lie within the boundary polygon and outside of all holes

        :param coords: numerical tuple representing coordinates
        :return: whether the given coordinate is a valid query point
        """
        return is_within_map(coords, self.boundary_polygon, self.holes)

    def get_visible_idxs(
        self,
        origin: int,
        candidates: Iterable[int],
        coords: np.ndarray,
        vert_idx2repr: np.ndarray,
        vert_idx2dist: np.ndarray,
    ) -> Set[int]:
        # Note: points with equal coordinates should not be considered visible (will be merged later)
        candidates = {i for i in candidates if not vert_idx2dist[i] == 0.0}
        edge_idxs2check = set(range(self.nr_edges))
        return find_visible(
            origin,
            candidates,
            edge_idxs2check,
            coords,
            vert_idx2repr,
            vert_idx2dist,
            self.extremity_mask,
            self.edge_vertex_idxs,
            self.vertex_edge_idxs,
        )

    def find_shortest_path(
        self,
        start_coordinates: InputCoords,
        goal_coordinates: InputCoords,
        free_space_after: bool = True,
        verify: bool = True,
    ) -> Tuple[PATH_TYPE, LENGTH_TYPE]:
        """computes the shortest path and its length between start and goal node

        :param start_coordinates: a (x,y) coordinate tuple representing the start node
        :param goal_coordinates:  a (x,y) coordinate tuple representing the goal node
        :param free_space_after: whether the created temporary search graph graph
            should be deleted after the query
        :param verify: whether it should be checked if start and goal points really lie inside the environment.
         if points close to or on polygon edges should be accepted as valid input, set this to ``False``.
        :return: a tuple of shortest path and its length. ([], None) if there is no possible path.
        """
        # path planning query:
        # make sure the map has been loaded and prepared
        if self.boundary_polygon is None:
            raise ValueError("No Polygons have been loaded into the map yet.")
        if not self.prepared:
            self.prepare()

        coords_start = np.array(start_coordinates, dtype=float)
        coords_goal = np.array(goal_coordinates, dtype=float)
        if verify and not self.within_map(coords_start):
            raise ValueError("start point does not lie within the map")
        if verify and not self.within_map(coords_goal):
            raise ValueError("goal point does not lie within the map")

        if np.array_equal(coords_start, coords_goal):
            # start and goal are identical and can be reached instantly
            return [start_coordinates, goal_coordinates], 0.0

        start = self.idx_start
        goal = self.idx_goal
        # temporarily extend data structure
        # Note: start and goal nodes could be identical with one ore more of the vertices
        # BUT: this is an edge case -> compute visibility as usual and later try to merge with the graph
        coords = np.append(self.coords, (coords_start, coords_goal), axis=0)
        self._coords_tmp = coords  # for plotting including the start and goal indices

        # check the goal node first (earlier termination possible)
        origin = goal
        # the visibility of only the graph nodes has to be checked (not all extremities!)
        # IMPORTANT: also check if the start node is visible from the goal node!
        candidate_idxs: Set[int] = set(self.graph.nodes)
        candidate_idxs.add(start)
        repr_n_dists = cmp_reps_n_distances(origin, coords)
        self.reprs_n_distances[origin] = repr_n_dists
        vert_idx2repr, vert_idx2dist = repr_n_dists
        visibles_goal = self.get_visible_idxs(origin, candidate_idxs, coords, vert_idx2repr, vert_idx2dist)
        if len(visibles_goal) == 0:
            # The goal node does not have any neighbours. Hence there is not possible path to the goal.
            return [], None

        # IMPORTANT geometrical property of this problem: it is always shortest to directly reach a node
        #   instead of visiting other nodes first (there is never an advantage through reduced edge weight)
        # -> when goal is directly reachable, there can be no other shorter path to it. Terminate
        if start in visibles_goal:
            d = vert_idx2dist[start]
            return [start_coordinates, goal_coordinates], d

        # create temporary graph
        # DirectedHeuristicGraph implements __deepcopy__() to not change the original precomputed self.graph
        # but to still not create real copies of vertex instances!
        graph = self.graph.copy()
        # TODO avoid real copy to make make more performant
        # graph = self.graph
        # nr_edges_before = len(graph.edges)

        # add unidirectional edges in the direction: extremity (v) -> goal
        for i in visibles_goal:
            graph.add_edge(i, goal, weight=vert_idx2dist[i])

        origin = start
        # the visibility of only the graphs nodes have to be checked
        # the goal node does not have to be considered, because of the earlier check
        repr_n_dists = cmp_reps_n_distances(origin, coords)
        self.reprs_n_distances[origin] = repr_n_dists
        vert_idx2repr, vert_idx2dist = repr_n_dists
        visibles_start = self.get_visible_idxs(origin, candidate_idxs, coords, vert_idx2repr, vert_idx2dist)

        if len(visibles_start) == 0:
            # The start node does not have any neighbours. Hence there is no possible path to the goal.
            return [], None

        # add edges in the direction: start -> extremity
        # Note: also here unnecessary edges in the graph could be deleted
        # optimising the graph here however is more expensive than beneficial,
        # as the graph is only being used for a single query
        for i in visibles_start:
            graph.add_edge(start, i, weight=vert_idx2dist[i])

        def l2_distance(n1, n2):
            return get_distance(n1, n2, self.reprs_n_distances)

        # apply mapping to start and goal index as well
        start_mapped = find_identical_single(start, graph.nodes, self.reprs_n_distances)
        if start_mapped != start:
            nx.relabel_nodes(graph, {start: start_mapped}, copy=False)

        goal_mapped = find_identical_single(goal, graph.nodes, self.reprs_n_distances)
        if goal_mapped != goal_mapped:
            nx.relabel_nodes(graph, {goal: goal_mapped}, copy=False)

        self._idx_start_tmp, self._idx_goal_tmp = start_mapped, goal_mapped  # for plotting

        id_path = nx.astar_path(graph, start_mapped, goal_mapped, heuristic=l2_distance, weight="weight")

        # clean up
        # TODO re-use the same graph. need to keep track of all merged edges
        # if start_mapped == start:
        #     graph.remove_node(start)
        # if goal_mapped==goal:
        #     graph.remove_node(goal)
        # nr_edges_after = len(graph.edges)
        # if not nr_edges_after == nr_edges_before:
        #     raise ValueError

        if free_space_after:
            del graph  # free the memory
        else:
            self.temp_graph = graph

        # compute distance
        distance = 0.0
        v1 = id_path[0]
        for v2 in id_path[1:]:
            distance += l2_distance(v1, v2)
            v1 = v2

        # extract the coordinates from the path
        path = [tuple(coords[i]) for i in id_path]
        return path, distance



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['extremitypathfinder']

package_data = \
{'': ['*']}

install_requires = \
['networkx>=2.8.5,<3.0.0', 'numpy>=1.22,<2.0']

entry_points = \
{'console_scripts': ['extremitypathfinder = '
                     'extremitypathfinder.command_line:main']}

setup_kwargs = {
    'name': 'extremitypathfinder',
    'version': '2.4.0',
    'description': 'python package implementing a multivariate Horner scheme for efficiently evaluating multivariate polynomials',
    'long_description': '===================\nextremitypathfinder\n===================\n\n\n.. image:: https://api.travis-ci.org/jannikmi/extremitypathfinder.svg?branch=master\n    :target: https://travis-ci.org/jannikmi/extremitypathfinder\n\n.. image:: https://readthedocs.org/projects/extremitypathfinder/badge/?version=latest\n    :alt: documentation status\n    :target: https://extremitypathfinder.readthedocs.io/en/latest/?badge=latest\n\n.. image:: https://img.shields.io/pypi/wheel/extremitypathfinder.svg\n    :target: https://pypi.python.org/pypi/extremitypathfinder\n\n.. image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n\n.. image:: https://pepy.tech/badge/extremitypathfinder\n    :alt: Total PyPI downloads\n    :target: https://pepy.tech/project/extremitypathfinder\n\n.. image:: https://img.shields.io/pypi/v/extremitypathfinder.svg\n    :alt: latest version on PyPI\n    :target: https://pypi.python.org/pypi/extremitypathfinder\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n\npython package for fast geometric shortest path computation in 2D multi-polygon or grid environments based on visibility graphs.\n\n\n.. image:: ./docs/_static/title_demo_plot.png\n\n\nQuick Guide:\n\n.. code-block:: console\n\n    pip install extremitypathfinder\n\n\n.. code-block:: python\n\n    from extremitypathfinder import PolygonEnvironment\n\n    environment = PolygonEnvironment()\n    # counter clockwise vertex numbering!\n    boundary_coordinates = [(0.0, 0.0), (10.0, 0.0), (9.0, 5.0), (10.0, 10.0), (0.0, 10.0)]\n    # clockwise numbering!\n    list_of_holes = [\n        [\n            (3.0, 7.0),\n            (5.0, 9.0),\n            (4.5, 7.0),\n            (5.0, 4.0),\n        ],\n    ]\n    environment.store(boundary_coordinates, list_of_holes, validate=False)\n    environment.prepare()\n    start_coordinates = (4.5, 1.0)\n    goal_coordinates = (4.0, 8.5)\n    path, length = environment.find_shortest_path(start_coordinates, goal_coordinates)\n\n\nFor more refer to the `documentation <https://extremitypathfinder.readthedocs.io/en/latest/>`__.\n\n\nAlso see:\n`GitHub <https://github.com/jannikmi/extremitypathfinder>`__,\n`PyPI <https://pypi.python.org/pypi/extremitypathfinder/>`__\n',
    'author': 'jannikmi',
    'author_email': 'github@michelfe.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://extremitypathfinder.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



def proto_test_case(data, fct):
    for input, expected_output in data:
        # print(input, expected_output, fct(input))
        actual_output = fct(input)
        if actual_output != expected_output:
            print("input: {} expected: {} got: {}".format(input, expected_output, actual_output))
        assert actual_output == expected_output
